/**
  Draw a green disk using an OpenGL shader.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2012-01-24 (Public Domain)
*/
#include "physics/world.h" /* physics::library and physics::object */
#include "ogl/glsl.h"
#include <fstream>
#include <sstream>

/* Just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"


class tri {
public:
	int verts[3];
};

/* A raytraced object */
class rayObject : public physics::object {
public:
	rayObject(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{ }
	
	void simulate(physics::library &lib) { }
	
	void draw(physics::library &lib) {
		/* Read the .obj file into this data structure: */
		static std::vector<vec3> verts; // 3D positions
		static std::vector<tri> tris; // vertex numbers
		if (verts.size()==0) {
			std::ifstream file("cat.obj");
			std::string line;
			while (std::getline(file,line)) {
				if (line[0]=='v') { /* vertex */
					vec3 v;
					std::stringstream ss(line+" ");
					char c; ss>>c; // skip the 'c'
					ss>>v.x>>v.y>>v.z;
					
					/* OBJ format has Y up, we want Z up */
					std::swap(v.y,v.z);
					v.y=-v.y;
					
					/* push model up vertically a tad */
					v.z+=0.2;
					if (!ss) {
						std::cout<<"Error in .obj v line!  What is "<<line<<"!?\n";
					}
					else 
						verts.push_back(v);
				}
				else if (line[0]=='f') { /* face: FIXME assumes triangle */
					tri t;
					std::stringstream ss(line+" ");
					char c; ss>>c; // skip the 'f'
					for (unsigned int j=0;j<3;j++)
					{
						ss>>t.verts[j];
						t.verts[j]--; // from 1-based obj file to 0-based truth!
					}
					if (!ss){
						std::cout<<"Error in .obj f line!  What is "<<line<<"!?\n";
					} else
						tris.push_back(t);
				}
			}
		}
		
		/* Draw the triangles */
		glBegin(GL_TRIANGLES);
		for (unsigned int i=0;i<tris.size();i++)
		{
			glColor3ub(rand()%256,rand()%256,rand()%256);
			for (unsigned int j=0;j<3;j++)
				glVertex3fv(verts[tris[i].verts[j]]);
		}
		glEnd();
		
		if (1) {
			glColor3f(0.0,0.0,0.0);
			/* draw vertices */
			glPointSize(4.0);
			glBegin(GL_POINTS);
			for (unsigned int i=0;i<verts.size();i++)
				glVertex3fv(verts[i]);
			glEnd();
		}
	
		/* Load up shader */
		GLhandleARB prog=MAKE_PROGRAM_OBJECT3(
		/* vertex program */
			void main(void) {
			   myColor = gl_Color;
			   /* Include everything in onscreen position */
			   gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;
			   /* Don't project, but do modelview transforms */
			   location=(vec3)(gl_ModelViewMatrix*gl_Vertex);
			}
		,
		/* shared vertex/fragment code */
			uniform vec3 camera; // world coordinates of camera
			varying vec4 myColor;
			varying vec3 location; // world coordinates of our pixel
		,
		/* fragment program */
			void main(void) {
				vec3 C=camera; // origin of ray
				vec3 D=location-camera; // direction of ray
			//	C.x-=1.0; // shift from world coords to object coords
			//	C.z*=1.9; // oblate spheroid
			//	D.z*=1.9;
				
			/* Quadratic: */
				float r=1.0;
				float a=dot(D,D);
				float b=2.0*dot(C,D);
				float c=dot(C,C)-r*r;
				float det=b*b-4.0*a*c;
				if (det<0) discard; /* miss: can't take square root of negative */
				float t=(-b-sqrt(det))/(2.0*a);
				
				
			/* Linear: 
				vec3 N=vec3(1,0,1); // normal of plane
				float a=dot(C,N);
				float b=dot(D,N);
				float t=-a/b;
			*/
				if (t<0) discard; //behind your head!
				vec3 P=C+t*D; // ray-object intersection point
				vec3 N=normalize(P); // object normal
				
				vec3 L=normalize(vec3(0.3,-0.5,1.0));
				vec3 H=normalize(L+normalize(-D));
				float specular=pow(dot(H,N),500.0);
				
				//if (length(P)>1.2) discard; // trim plane 
				gl_FragColor.rgb=dot(N,L)*vec3(0.8,0.0,0.2)+specular*vec3(1,1,1);
				gl_FragColor.a=1.0; // opaque
				
			/* Compute depth */
				vec4 projP = gl_ProjectionMatrix*vec4(P,1);
				float depth = projP.z/projP.w; // perspective divide
				gl_FragDepth = 0.5+0.5 * depth; // compensate for glDepthRange
			}
		);
		
		glUseProgramObjectARB(prog);
		unsigned int ul=glGetUniformLocationARB(prog,"camera");
		glUniform3fvARB(ul,1,camera);

		/* Draw a big green disk */
		glColor4f(0.0f,1.0f,0.0f,0.8f);
		
		glutSolidSphere(1.5,4,8);
		/*
		glBegin(GL_QUADS);
		glVertex3f(-1e3,0,-1e3);
		glVertex3f(-1e3,0,+1e3);
		glVertex3f(+1e3,0,+1e3);
		glVertex3f(+1e3,0,-1e3);
		glEnd();
		*/
		
		glUseProgramObjectARB(0);
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new rayObject());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs usual alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Draw opaque white ground plane */
	float groundz=-1.0;
	float groundsize=30;
	glEnable(GL_DEPTH_TEST);
	glColor3f(0.2,0.4,0.6);
	glBegin(GL_QUADS);
	glVertex3f(-groundsize,-groundsize,groundz);
	glVertex3f(+groundsize,-groundsize,groundz);
	glVertex3f(+groundsize,+groundsize,groundz);
	glVertex3f(-groundsize,+groundsize,groundz);
	glEnd();
	
	/* No depth while drawing ground grid */
	glDisable(GL_DEPTH_TEST);
	glColor4f(0.0,0.0,0.0,1.0);
	glEnable(GL_LINE_SMOOTH);
	groundz-=0.001;
	glLineWidth(2.0);
	if (camera.z>=groundz) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (float grid=-groundsize;grid<=+groundsize;grid++) {
			glVertex3f(-groundsize,grid,groundz); 
			glVertex3f(+groundsize,grid,groundz); 
			glVertex3f(grid,-groundsize,groundz); 
			glVertex3f(grid,+groundsize,groundz); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
	
	
}


